@@ -1,5 +1,8 @@
 #!/usr/bin/env python
 
+from __future__ import print_function
+from builtins import bytes
+from builtins import object
 from builtins import next
 from builtins import str
 
@@ -7,17 +10,16 @@ import click
 import json
 import locale
 import os
-import requests
 import time
 import sys
 import xml.etree.ElementTree as ET
 import zipfile
 
-from transcriptic import analyze as api_analyze, submit as api_submit
 from transcriptic.english import AutoprotocolParser
 from transcriptic.config import Connection
 from transcriptic.objects import ProtocolPreview
 from transcriptic.util import iter_json
+from transcriptic import routes
 from os.path import isfile
 from collections import OrderedDict
 from contextlib import contextmanager
@@ -30,6 +32,20 @@ except NameError:
     pass
 
 
+class ContextObject(object):
+    """Object passed along Click context"""
+    def __init__(self):
+        self._api = None
+
+    @property
+    def api(self):
+        return self._api
+
+    @api.setter
+    def api(self, value):
+        self._api = value
+
+
 @click.group()
 @click.option('--apiroot', default=None)
 @click.option(
@@ -39,16 +55,22 @@ except NameError:
     help='Specify a configuration file.'
 )
 @click.option('--organization', '-o', default=None)
+@click.version_option(prog_name="Transcriptic Python Library (TxPy)")
 @click.pass_context
 def cli(ctx, apiroot, config, organization):
     """A command line tool for working with Transcriptic."""
-    if ctx.invoked_subcommand not in ['login']:
+    if ctx.invoked_subcommand in ['login']:
+        # Initialize empty connection
+        ctx.obj = ContextObject()
+        ctx.obj.api = Connection(use_environ=False)
+    elif ctx.invoked_subcommand not in ['compile', 'preview', 'summarize', 'init']:
         try:
-            ctx.obj = Connection.from_file(config)
+            ctx.obj = ContextObject()
+            ctx.obj.api = Connection.from_file(config)
             if organization is not None:
-                ctx.obj.organization_id = organization
+                ctx.obj.api.organization_id = organization
             if apiroot is not None:
-                ctx.obj.api_root = apiroot
+                ctx.obj.api.api_root = apiroot
         except IOError:
             click.echo("Error reading config file, running "
                        "`transcriptic login` ...")
@@ -81,10 +103,10 @@ def submit(ctx, file, project, title, test):
             return
 
     try:
-        req_json = api_submit(protocol, project, title, test_mode=test)
+        req_json = ctx.obj.api.submit_run(protocol, project_id=project, title=title, test_mode=test)
         run_id = req_json['id']
         click.echo("Run created: %s" %
-                   ctx.obj.url("%s/runs/%s" % (project, run_id)))
+                   ctx.obj.api.url("%s/runs/%s" % (project, run_id)))
     except Exception as e:
         click.echo(str(e))
 
@@ -148,19 +170,9 @@ def upload_release(ctx, archive, package):
                            show_eta=False, width=70,
                            fill_char="|", empty_char="-") as bar:
         bar.update(10)
-        sign = requests.get(
-            'https://secure.transcriptic.com/upload/sign',
-            params={'name': archive},
-            headers={
-                'X-User-Email': ctx.obj.email,
-                'X-User-Token': ctx.obj.token,
-                'Content-Type': 'application/json',
-                'Accept': 'application/json',
-                }
-            )
-        info = json.loads(sign.content)
+        info = ctx.obj.api.get(ctx.obj.api.get_route('upload_sign'), params={'name': archive})
         bar.update(30)
-        url = 'https://transcriptic-uploads.s3.amazonaws.com'
+        aws_url = ctx.obj.api.get_route('aws_upload')
         files = {'file': open(os.path.basename(archive), 'rb')}
         data = OrderedDict([
             ('key', info['key']),
@@ -170,21 +182,19 @@ def upload_release(ctx, archive, package):
             ('policy', info['policy']),
             ('signature', info['signature']),
         ])
-        response = requests.post(url, data=data, files=files)
+        response = ctx.obj.api.post(aws_url, data=data, files=files, headers={},
+                            status_response={'201': lambda resp: resp})
         bar.update(20)
         response_tree = ET.fromstring(response.content)
         loc = dict((i.tag, i.text) for i in response_tree)
         try:
-            up = ctx.obj.post(
-                'packages/%s/releases/' % package_id,
+            up = ctx.obj.api.post_release(
                 data=json.dumps({"release":
                                 {"binary_attachment_url": loc["Key"]}}
                                 ),
-                headers={
-                    "Origin": "https://secure.transcriptic.com/",
-                    "Content-Type": "application/json"}
+                package_id=package_id
                 )
-            re = json.loads(up.content)['id']
+            re = up['id']
         except ValueError:
             click.echo("\nError: There was a problem uploading your release."
                        "\nVerify that your manifest.json file is properly  "
@@ -195,12 +205,10 @@ def upload_release(ctx, archive, package):
             return
         bar.update(20)
         time.sleep(10)
-        status = ctx.obj.get(
-            'packages/%s/releases/%s?_=%s' %
-            (package_id, re, int(time.time()))
-            )
-        published = json.loads(status.content)['published']
-        errors = status.json()['validation_errors']
+        status = ctx.obj.api.get_release_status(package_id=package_id, release_id=re,
+                                            timestamp=int(time.time()))
+        published = status['published']
+        errors = status['validation_errors']
         bar.update(30)
         if errors:
             click.echo("\nPackage upload to %s unsuccessful. "
@@ -211,7 +219,7 @@ def upload_release(ctx, archive, package):
                        )
         else:
             click.echo("\nPackage uploaded successfully! \n"
-                       "Visit %s to publish." % ctx.obj.url('packages/%s' %
+                       "Visit %s to publish." % ctx.obj.api.url('packages/%s' %
                                                             package_id))
 
 
@@ -239,26 +247,26 @@ def protocols():
 @click.option("-i")
 def packages(ctx, i):
     """List packages in your organization."""
-    response = ctx.obj.get('packages/')
+    response = ctx.obj.api.packages()
     # there's probably a better way to do this
     package_names = OrderedDict(
         sorted(list({"yours": {}, "theirs": {}}.items()),
                key=lambda t: len(t[0]))
         )
-    if response.status_code == 200:
-        for pack in response.json():
-            n = str(pack['name']).lower().replace(
-                "com.%s." % ctx.obj.organization_id, "")
-            latest = str(pack['latest_version']) if pack[
-                'latest_version'] else "-"
-            if pack.get('owner') and pack['owner']['email'] == ctx.obj.email:
-                package_names['yours'][n] = {}
-                package_names['yours'][n]['id'] = str(pack['id'])
-                package_names['yours'][n]['latest'] = latest
-            else:
-                package_names['theirs'][n] = {}
-                package_names['theirs'][n]['id'] = str(pack['id'])
-                package_names['theirs'][n]['latest'] = latest
+
+    for pack in response:
+        n = str(pack['name']).lower().replace(
+            "com.%s." % ctx.obj.api.organization_id, "")
+        latest = str(pack['latest_version']) if pack[
+            'latest_version'] else "-"
+        if pack.get('owner') and pack['owner']['email'] == ctx.obj.api.email:
+            package_names['yours'][n] = {}
+            package_names['yours'][n]['id'] = str(pack['id'])
+            package_names['yours'][n]['latest'] = latest
+        else:
+            package_names['theirs'][n] = {}
+            package_names['theirs'][n]['id'] = str(pack['id'])
+            package_names['theirs'][n]['latest'] = latest
     if i:
         return dict(list(package_names['yours'].items()) +
                     list(package_names['theirs'].items()))
@@ -289,22 +297,19 @@ def packages(ctx, i):
 @click.pass_context
 def create_package(ctx, description, name):
     """Create a new empty protocol package"""
-    existing = ctx.obj.get('packages/')
-    for p in existing.json():
+    existing = ctx.obj.api.packages()
+    for p in existing:
         if name == p['name'].split('.')[-1]:
             click.echo("You already have an existing package with the name "
                        "\"%s\". Please choose a different package name." %
                        name)
             return
-    new_pack = ctx.obj.post('packages/', data=json.dumps({
-        "description": description,
-        "name": "%s%s" % ("com.%s." % ctx.obj.organization_id, name)
-    }))
-    if new_pack.status_code == 201:
+    new_pack = ctx.obj.api.create_package(name, description)
+    if new_pack:
         click.echo("New package '%s' created with id %s \n"
-                   "View it at %s" % (name, new_pack.json()['id'],
-                                      ctx.obj.url('packages/%s' %
-                                                  new_pack.json()['id'])
+                   "View it at %s" % (name, new_pack['id'],
+                                      ctx.obj.api.url('packages/%s' %
+                                                  new_pack['id'])
                                       )
                    )
     else:
@@ -318,16 +323,16 @@ def create_package(ctx, description, name):
 @click.pass_context
 def delete_package(ctx, name, force):
     """Delete an existing protocol package"""
-    id = get_package_id(name)
-    if id:
+    package_id = get_package_id(name)
+    if package_id:
         if not force:
             click.confirm(
                 "Are you sure you want to permanently delete the package "
                 "'%s'?  All releases within will be lost." %
-                get_package_name(id), default=False, abort=True
+                get_package_name(package_id), default=False, abort=True
             )
             click.confirm("Are you really really sure?", default=True)
-        del_pack = ctx.obj.delete_package(id)
+        del_pack = ctx.obj.api.delete_package(package_id=package_id)
         if del_pack:
             click.echo("Package deleted.")
         else:
@@ -340,13 +345,13 @@ def delete_package(ctx, name, force):
 def projects(ctx, i):
     """List the projects in your organization"""
     try:
-        projects = ctx.obj.projects()
+        projects = ctx.obj.api.projects()
         proj_names = {}
         all_proj = {}
         for proj in projects:
-            status = " (archived)" if proj.attributes['archived_at'] else ""
-            proj_names[proj.attributes['name']] = proj.id
-            all_proj[proj.attributes['name'] + status] = proj.id
+            status = " (archived)" if proj['archived_at'] else ""
+            proj_names[proj["name"]] = proj["id"]
+            all_proj[proj["name"] + status] = proj["id"]
         if i:
             return proj_names
         else:
@@ -368,19 +373,19 @@ def projects(ctx, i):
 @click.argument('project_name')
 def runs(ctx, project_name):
     """List the runs that exist in a project"""
-    id = get_project_id(project_name)
-    runs = []
-    if id:
-        req = ctx.obj.runs(id)
-        if not req['runs']:
+    project_id = get_project_id(project_name)
+    run_list = []
+    if project_id:
+        req = ctx.obj.api.runs(project_id=project_id)
+        if not req:
             click.echo("Project '%s' is empty." % project_name)
             return
-        for r in req['runs']:
-            runs.append([r['title'] or "(Untitled)",
-                         r['id'],
-                         r['completed_at'].split("T")[0] if r['completed_at']
-                         else r['created_at'].split("T")[0],
-                         r['status'].replace("_", " ")])
+        for r in req:
+            run_list.append([r['title'] or "(Untitled)",
+                             r['id'],
+                             r['completed_at'].split("T")[0] if r['completed_at']
+                             else r['created_at'].split("T")[0],
+                             r['status'].replace("_", " ")])
 
         click.echo(
             '\n{:^120}'.format("Runs in Project '%s':\n" %
@@ -391,7 +396,7 @@ def runs(ctx, project_name):
                    '{:^30}'.format("RUN DATE") + "|" +
                    '{:^30}'.format('RUN STATUS'))
         click.echo('{:-^120}'.format(''))
-        for run in runs:
+        for run in run_list:
             click.echo('{:^30}'.format(run[0]) + "|" +
                        '{:^30}'.format(run[1]) + "|" +
                        '{:^30}'.format(run[2]) + "|" +
@@ -406,19 +411,18 @@ def runs(ctx, project_name):
 @click.pass_context
 def create_project(ctx, name, dev):
     """Create a new empty project."""
-    existing = ctx.obj.projects()
+    existing = ctx.obj.api.projects()
     for p in existing:
-        if name == p.attributes['name'].split('.')[-1]:
+        if name == p['name'].split('.')[-1]:
             click.echo("You already have an existing project with the name "
                        "\"%s\".  Please choose a different project name." %
                        name)
             return
     try:
-        new_proj = ctx.obj.create_project(name)
+        new_proj = ctx.obj.api.create_project(name)
         click.echo(
             "New%s project '%s' created with id %s  \nView it at %s" % (
-                " pilot" if dev else "", name, new_proj.attributes[
-                    'id'], ctx.obj.url('%s' % (new_proj.attributes['id']))
+                " pilot" if dev else "", name, new_proj['id'], ctx.obj.api.url('%s' % (new_proj['id']))
             )
         )
     except RuntimeError:
@@ -432,16 +436,16 @@ def create_project(ctx, name, dev):
 @click.pass_context
 def delete_project(ctx, name, force):
     """Delete an existing project."""
-    id = get_project_id(name)
-    if id:
+    project_id = get_project_id(name)
+    if project_id:
         if not force:
             click.confirm(
                 "Are you sure you want to permanently delete '%s'?" %
-                get_project_name(id),
+                get_project_name(project_id),
                 default=False,
                 abort=True
             )
-        if ctx.obj.delete_project(str(id)):
+        if ctx.obj.api.delete_project(project_id=str(project_id)):
             click.echo("Project deleted.")
         else:
             click.confirm(
@@ -450,7 +454,7 @@ def delete_project(ctx, name, force):
                 default=False,
                 abort=True
             )
-            if ctx.obj.archive_project(str(id)):
+            if ctx.obj.api.archive_project(project_id=str(project_id)):
                 click.echo("Project archived.")
             else:
                 click.echo("Could not archive project!")
@@ -461,7 +465,7 @@ def delete_project(ctx, name, force):
 @click.pass_context
 def resources(ctx, query):
     """Search catalog of provisionable resources"""
-    req = ctx.obj.resources(query)
+    req = ctx.obj.api.resources(query)
     if req["results"]:
         click.echo("Results for '%s':" % query)
         click.echo('{:^40}'.format("Resource Name") + '|' +
@@ -473,12 +477,12 @@ def resources(ctx, query):
     for i in req["results"]:
         if i["provisionable"] and not i["reservable"]:
             name = i['name'].encode('ascii', errors='ignore')
-            id = i['kit_items'][0]['resource_id']
+            resource_id = i['kit_items'][0]['resource_id']
             click.echo('{:^40}'.format(name) + '|' +
                        '{:^40}'.format(i['vendor']['name'] if 'vendor' in
                                        list(i.keys()) else ''
                                        ) +
-                       '|' + '{:^40}'.format(id))
+                       '|' + '{:^40}'.format(resource_id))
             click.echo('{:-^120}'.format(''))
 
 
@@ -533,7 +537,7 @@ def analyze(ctx, file, test):
             return
 
     try:
-        analysis = api_analyze(protocol, test_mode=test)
+        analysis = ctx.obj.api.analyze_run(protocol, test_mode=test)
         click.echo(u"\u2713 Protocol analyzed")
         price(analysis)
     except Exception as e:
@@ -639,14 +643,11 @@ def launch(ctx, protocol, project, save_input):
 
     manifest, protocol = load_manifest_and_protocol(protocol)
 
-    res = ctx.obj.post('%s/runs/quick_launch' %
-                       project, data=json.dumps({"manifest": protocol}))
-    quick_launch = res.json()
+    quick_launch = ctx.obj.api.create_quick_launch(data=json.dumps({"manifest": protocol}), project_id=project)
     quick_launch_mtime = quick_launch["updated_at"]
 
     format_str = "\nOpening %s"
-    url = ctx.obj.url("%s/runs/quick_launch/%s" %
-                      (project, quick_launch["id"]))
+    url = ctx.obj.api.get_route('get_quick_launch', project_id=project, quick_launch_id=quick_launch["id"])
     print_stderr(format_str % url)
 
     """
@@ -676,9 +677,8 @@ def launch(ctx, protocol, project, save_input):
         sys.stderr.flush()
         time.sleep(5)
 
-        res = ctx.obj.get('%s/runs/quick_launch/%s' %
-                          (project, quick_launch["id"]))
-        quick_launch = res.json()
+        quick_launch = ctx.obj.api.get_quick_launch(project_id=project,
+                                                quick_launch_id=quick_launch["id"])
         count += 1
 
     # Save the protocol input locally if the user specified the save_input
@@ -701,7 +701,7 @@ def login(ctx, api_root):
     """Authenticate to your Transcriptic account."""
     email = click.prompt('Email')
     password = click.prompt('Password', hide_input=True)
-    r = requests.post("%s/users/sign_in" % api_root, data=json.dumps({
+    r = ctx.obj.api.post(routes.login(api_root=api_root), data=json.dumps({
         'user': {
             'email': email,
             'password': password,
@@ -709,7 +709,11 @@ def login(ctx, api_root):
     }), headers={
         'Accept': 'application/json',
         'Content-Type': 'application/json',
-    })
+    }, status_response = {
+        '200': lambda resp: resp,
+        'default': lambda resp: resp
+    },
+     custom_request=False)
     if r.status_code != 200:
         click.echo("Error logging into Transcriptic: %s" % r.json()['error'])
         sys.exit(1)
@@ -743,30 +747,34 @@ def login(ctx, api_root):
             value_proc=lambda x: parse_valid_org(x)
         )
 
-    r = requests.get('%s/%s' % (api_root, organization), headers={
+    r = ctx.obj.api.get(routes.get_organizations(api_root=api_root, org_id=organization), headers={
         'X-User-Email': email,
         'X-User-Token': token,
         'Accept': 'application/json',
-    })
+    }, status_response={
+        '200': lambda resp: resp,
+        'default': lambda resp: resp
+    },
+        custom_request=True)
     if r.status_code != 200:
         click.echo("Error accessing organization: %s" % r.text)
         sys.exit(1)
-    ctx.obj = Connection(email, token, organization, api_root=api_root)
-    ctx.obj.save(ctx.parent.params['config'])
+    ctx.obj.api = Connection(email=email, token=token, organization_id=organization, api_root=api_root)
+    ctx.obj.api.save(ctx.parent.params['config'])
     click.echo('Logged in as %s (%s)' % (user['email'], organization))
 
 
 @click.pass_context
 def get_project_id(ctx, name):
     projs = ctx.invoke(projects, i=True)
-    id = projs.get(name)
-    if not id:
-        id = name if name in list(projs.values()) else None
-        if not id:
+    project_id = projs.get(name)
+    if not project_id:
+        project_id = name if name in list(projs.values()) else None
+        if not project_id:
             click.echo(
                 "The project '%s' was not found in your organization." % name)
             return
-    return id
+    return project_id
 
 
 @click.pass_context
@@ -798,15 +806,15 @@ def get_package_id(ctx, name):
 
 
 @click.pass_context
-def get_package_name(ctx, id):
+def get_package_name(ctx, package_id):
     package_names = {
         v['id']: k for k, v in list(ctx.invoke(packages, i=True).items())}
-    package_name = package_names.get(id)
+    package_name = package_names.get(package_id)
     if not package_name:
-        package_name = id if id in list(package_names.values()) else None
+        package_name = package_id if package_id in list(package_names.values()) else None
     if not package_name:
         click.echo("The id '%s' does not match any package in your \
-                    organization." % id)
+                    organization." % package_id)
         return
     return package_name
 
@@ -865,15 +873,15 @@ def run_protocol(ctx, manifest, protocol, inputs, view=False):
     from subprocess import check_output, CalledProcessError
     import tempfile
     with tempfile.NamedTemporaryFile() as fp:
-        fp.write(json.dumps(inputs))
+        fp.write(bytes(json.dumps(inputs), 'UTF-8'))
         fp.flush()
         try:
             protocol = check_output(["bash", "-c", command + " " + fp.name])
             click.echo(protocol)
             if view:
                 click.echo("View your protocol's raw JSON above or see the "
-                           "intructions rendered at the following link: \n%s" %
-                           ProtocolPreview(protocol, ctx.obj).preview_url)
+                           "instructions rendered at the following link: \n%s" %
+                           ProtocolPreview(protocol, ctx.obj.api).preview_url)
         except CalledProcessError as e:
             click.echo(e.output)
             return
